@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-#!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 #***************************************************************************
 #                                  _   _ ____  _
@@ -33,6 +32,7 @@ from datetime import timedelta, datetime
 from json import JSONEncoder
 import time
 from typing import List, Union, Optional
+import copy
 
 from .curl import CurlClient, ExecResult
 from .env import Env
@@ -48,7 +48,7 @@ class Httpd:
         'authn_core', 'authn_file',
         'authz_user', 'authz_core', 'authz_host',
         'auth_basic', 'auth_digest',
-        'alias', 'env', 'filter', 'headers', 'mime', 'setenvif',
+        'alias', 'env', 'filter', 'headers', 'mime', 'setenvif', 'negotiation',
         'socache_shmcb',
         'rewrite', 'http2', 'ssl', 'proxy', 'proxy_http', 'proxy_connect',
         'brotli',
@@ -63,7 +63,6 @@ class Httpd:
 
     def __init__(self, env: Env, proxy_auth: bool = False):
         self.env = env
-        self._cmd = env.apachectl
         self._apache_dir = os.path.join(env.gen_dir, 'apache')
         self._run_dir = os.path.join(self._apache_dir, 'run')
         self._lock_dir = os.path.join(self._apache_dir, 'locks')
@@ -79,6 +78,7 @@ class Httpd:
         self._auth_digest = True
         self._proxy_auth_basic = proxy_auth
         self._extra_configs = {}
+        self._loaded_extra_configs = None
         assert env.apxs
         p = subprocess.run(args=[env.apxs, '-q', 'libexecdir'],
                            capture_output=True, text=True)
@@ -86,7 +86,7 @@ class Httpd:
             raise Exception(f'{env.apxs} failed to query libexecdir: {p}')
         self._mods_dir = p.stdout.strip()
         if self._mods_dir is None:
-            raise Exception(f'apache modules dir cannot be found')
+            raise Exception('apache modules dir cannot be found')
         if not os.path.exists(self._mods_dir):
             raise Exception(f'apache modules dir does not exist: {self._mods_dir}')
         self._process = None
@@ -101,7 +101,7 @@ class Httpd:
         self._rmf(self._error_log)
 
     def exists(self):
-        return os.path.exists(self._cmd)
+        return os.path.exists(self.env.httpd)
 
     def set_extra_config(self, domain: str, lines: Optional[Union[str, List[str]]]):
         if lines is None:
@@ -116,9 +116,7 @@ class Httpd:
         self._proxy_auth_basic = active
 
     def _run(self, args, intext=''):
-        env = {}
-        for key, val in os.environ.items():
-            env[key] = val
+        env = os.environ.copy()
         env['APACHE_RUN_DIR'] = self._run_dir
         env['APACHE_RUN_USER'] = os.environ['USER']
         env['APACHE_LOCK_DIR'] = self._lock_dir
@@ -133,8 +131,8 @@ class Httpd:
                           stderr=p.stderr.decode().splitlines(),
                           duration=datetime.now() - start)
 
-    def _apachectl(self, cmd: str):
-        args = [self.env.apachectl,
+    def _cmd_httpd(self, cmd: str):
+        args = [self.env.httpd,
                 "-d", self._apache_dir,
                 "-f", self._conf_file,
                 "-k", cmd]
@@ -148,14 +146,16 @@ class Httpd:
             fd.write('start of server\n')
         with open(os.path.join(self._apache_dir, 'xxx'), 'a') as fd:
             fd.write('start of server\n')
-        r = self._apachectl('start')
+        r = self._cmd_httpd('start')
         if r.exit_code != 0:
             log.error(f'failed to start httpd: {r}')
             return False
+        self._loaded_extra_configs = copy.deepcopy(self._extra_configs)
         return self.wait_live(timeout=timedelta(seconds=5))
 
     def stop(self):
-        r = self._apachectl('stop')
+        r = self._cmd_httpd('stop')
+        self._loaded_extra_configs = None
         if r.exit_code == 0:
             return self.wait_dead(timeout=timedelta(seconds=5))
         log.fatal(f'stopping httpd failed: {r}')
@@ -167,11 +167,18 @@ class Httpd:
 
     def reload(self):
         self._write_config()
-        r = self._apachectl("graceful")
+        r = self._cmd_httpd("graceful")
+        self._loaded_extra_configs = None
         if r.exit_code != 0:
             log.error(f'failed to reload httpd: {r}')
+        self._loaded_extra_configs = copy.deepcopy(self._extra_configs)
         return self.wait_live(timeout=timedelta(seconds=5))
 
+    def reload_if_config_changed(self):
+        if self._loaded_extra_configs == self._extra_configs:
+            return True
+        return self.reload()
+
     def wait_dead(self, timeout: timedelta):
         curl = CurlClient(env=self.env, run_dir=self._tmp_dir)
         try_until = datetime.now() + timeout
@@ -207,10 +214,16 @@ class Httpd:
         domain1 = self.env.domain1
         domain1brotli = self.env.domain1brotli
         creds1 = self.env.get_credentials(domain1)
+        assert creds1  # convince pytype this isn't None
         domain2 = self.env.domain2
         creds2 = self.env.get_credentials(domain2)
+        assert creds2  # convince pytype this isn't None
+        exp_domain = self.env.expired_domain
+        exp_creds = self.env.get_credentials(exp_domain)
+        assert exp_creds  # convince pytype this isn't None
         proxy_domain = self.env.proxy_domain
         proxy_creds = self.env.get_credentials(proxy_domain)
+        assert proxy_creds  # convince pytype this isn't None
         self._mkpath(self._conf_dir)
         self._mkpath(self._logs_dir)
         self._mkpath(self._tmp_dir)
@@ -239,52 +252,47 @@ class Httpd:
                 if os.path.exists(os.path.join(self._mods_dir, f'mod_{m}.so')):
                     fd.write(f'LoadModule {m}_module   "{self._mods_dir}/mod_{m}.so"\n')
             if Httpd.MOD_CURLTEST is not None:
-                fd.write(f'LoadModule curltest_module   \"{Httpd.MOD_CURLTEST}\"\n')
+                fd.write(f'LoadModule curltest_module   "{Httpd.MOD_CURLTEST}"\n')
             conf = [   # base server config
                 f'ServerRoot "{self._apache_dir}"',
-                f'DefaultRuntimeDir logs',
-                f'PidFile httpd.pid',
+                'DefaultRuntimeDir logs',
+                'PidFile httpd.pid',
                 f'ErrorLog {self._error_log}',
                 f'LogLevel {self._get_log_level()}',
-                f'StartServers 4',
-                f'ReadBufferSize 16000',
-                f'H2MinWorkers 16',
-                f'H2MaxWorkers 256',
-                f'H2Direct on',
+                'StartServers 4',
+                'ReadBufferSize 16000',
+                'H2MinWorkers 16',
+                'H2MaxWorkers 256',
                 f'Listen {self.env.http_port}',
                 f'Listen {self.env.https_port}',
                 f'Listen {self.env.proxy_port}',
                 f'Listen {self.env.proxys_port}',
                 f'TypesConfig "{self._conf_dir}/mime.types',
-                f'SSLSessionCache "shmcb:ssl_gcache_data(32000)"',
-                (f'SSLCipherSuite SSL'
-                 f' ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256'
-                 f':ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305'
-                ),
-                (f'SSLCipherSuite TLSv1.3'
-                 f' TLS_AES_128_GCM_SHA256:TLS_CHACHA20_POLY1305_SHA256'
-                ),
+                'SSLSessionCache "shmcb:ssl_gcache_data(32000)"',
+                'AddEncoding x-gzip .gz .tgz .gzip',
+                'AddHandler type-map .var',
             ]
             if 'base' in self._extra_configs:
                 conf.extend(self._extra_configs['base'])
             conf.extend([  # plain http host for domain1
                 f'<VirtualHost *:{self.env.http_port}>',
                 f'    ServerName {domain1}',
-                f'    ServerAlias localhost',
+                '    ServerAlias localhost',
                 f'    DocumentRoot "{self._docs_dir}"',
-                f'    Protocols h2c http/1.1',
+                '    Protocols h2c http/1.1',
+                '    H2Direct on',
             ])
             conf.extend(self._curltest_conf(domain1))
             conf.extend([
-                f'</VirtualHost>',
-                f'',
+                '</VirtualHost>',
+                '',
             ])
             conf.extend([  # https host for domain1, h1 + h2
                 f'<VirtualHost *:{self.env.https_port}>',
                 f'    ServerName {domain1}',
-                f'    ServerAlias localhost',
-                f'    Protocols h2 http/1.1',
-                f'    SSLEngine on',
+                '    ServerAlias localhost',
+                '    Protocols h2 http/1.1',
+                '    SSLEngine on',
                 f'    SSLCertificateFile {creds1.cert_file}',
                 f'    SSLCertificateKeyFile {creds1.pkey_file}',
                 f'    DocumentRoot "{self._docs_dir}"',
@@ -293,44 +301,44 @@ class Httpd:
             if domain1 in self._extra_configs:
                 conf.extend(self._extra_configs[domain1])
             conf.extend([
-                f'</VirtualHost>',
-                f'',
+                '</VirtualHost>',
+                '',
             ])
             # Alternate to domain1 with BROTLI compression
             conf.extend([  # https host for domain1, h1 + h2
                 f'<VirtualHost *:{self.env.https_port}>',
                 f'    ServerName {domain1brotli}',
-                f'    Protocols h2 http/1.1',
-                f'    SSLEngine on',
+                '    Protocols h2 http/1.1',
+                '    SSLEngine on',
                 f'    SSLCertificateFile {creds1.cert_file}',
                 f'    SSLCertificateKeyFile {creds1.pkey_file}',
                 f'    DocumentRoot "{self._docs_dir}"',
-                f'    SetOutputFilter BROTLI_COMPRESS',
+                '    SetOutputFilter BROTLI_COMPRESS',
             ])
             conf.extend(self._curltest_conf(domain1))
             if domain1 in self._extra_configs:
                 conf.extend(self._extra_configs[domain1])
             conf.extend([
-                f'</VirtualHost>',
-                f'',
+                '</VirtualHost>',
+                '',
             ])
             conf.extend([  # plain http host for domain2
                 f'<VirtualHost *:{self.env.http_port}>',
                 f'    ServerName {domain2}',
-                f'    ServerAlias localhost',
+                '    ServerAlias localhost',
                 f'    DocumentRoot "{self._docs_dir}"',
-                f'    Protocols h2c http/1.1',
+                '    Protocols h2c http/1.1',
             ])
             conf.extend(self._curltest_conf(domain2))
             conf.extend([
-                f'</VirtualHost>',
-                f'',
+                '</VirtualHost>',
+                '',
             ])
             conf.extend([  # https host for domain2, no h2
                 f'<VirtualHost *:{self.env.https_port}>',
                 f'    ServerName {domain2}',
-                f'    Protocols http/1.1',
-                f'    SSLEngine on',
+                '    Protocols http/1.1',
+                '    SSLEngine on',
                 f'    SSLCertificateFile {creds2.cert_file}',
                 f'    SSLCertificateKeyFile {creds2.pkey_file}',
                 f'    DocumentRoot "{self._docs_dir}/two"',
@@ -339,65 +347,84 @@ class Httpd:
             if domain2 in self._extra_configs:
                 conf.extend(self._extra_configs[domain2])
             conf.extend([
-                f'</VirtualHost>',
-                f'',
+                '</VirtualHost>',
+                '',
+            ])
+            conf.extend([  # https host for expired domain
+                f'<VirtualHost *:{self.env.https_port}>',
+                f'    ServerName {exp_domain}',
+                '    Protocols h2 http/1.1',
+                '    SSLEngine on',
+                f'    SSLCertificateFile {exp_creds.cert_file}',
+                f'    SSLCertificateKeyFile {exp_creds.pkey_file}',
+                f'    DocumentRoot "{self._docs_dir}/expired"',
+            ])
+            conf.extend(self._curltest_conf(exp_domain))
+            if exp_domain in self._extra_configs:
+                conf.extend(self._extra_configs[exp_domain])
+            conf.extend([
+                '</VirtualHost>',
+                '',
             ])
             conf.extend([  # http forward proxy
                 f'<VirtualHost *:{self.env.proxy_port}>',
                 f'    ServerName {proxy_domain}',
-                f'    Protocols h2c http/1.1',
-                f'    ProxyRequests On',
-                f'    H2ProxyRequests On',
-                f'    ProxyVia On',
+                '    Protocols h2c http/1.1',
+                '    ProxyRequests On',
+                '    H2ProxyRequests On',
+                '    ProxyVia On',
                 f'    AllowCONNECT {self.env.http_port} {self.env.https_port}',
             ])
             conf.extend(self._get_proxy_conf())
             conf.extend([
-                f'</VirtualHost>',
-                f'',
+                '</VirtualHost>',
+                '',
             ])
             conf.extend([  # https forward proxy
                 f'<VirtualHost *:{self.env.proxys_port}>',
                 f'    ServerName {proxy_domain}',
-                f'    Protocols h2 http/1.1',
-                f'    SSLEngine on',
+                '    Protocols h2 http/1.1',
+                '    SSLEngine on',
                 f'    SSLCertificateFile {proxy_creds.cert_file}',
                 f'    SSLCertificateKeyFile {proxy_creds.pkey_file}',
-                f'    ProxyRequests On',
-                f'    H2ProxyRequests On',
-                f'    ProxyVia On',
+                '    ProxyRequests On',
+                '    H2ProxyRequests On',
+                '    ProxyVia On',
                 f'    AllowCONNECT {self.env.http_port} {self.env.https_port}',
             ])
             conf.extend(self._get_proxy_conf())
             conf.extend([
-                f'</VirtualHost>',
-                f'',
+                '</VirtualHost>',
+                '',
             ])
 
             fd.write("\n".join(conf))
         with open(os.path.join(self._conf_dir, 'mime.types'), 'w') as fd:
             fd.write("\n".join([
+                'text/plain            txt',
                 'text/html             html',
                 'application/json      json',
+                'application/x-gzip    gzip',
+                'application/x-gzip    gz',
                 ''
             ]))
 
     def _get_proxy_conf(self):
         if self._proxy_auth_basic:
             return [
-                f'    <Proxy "*">',
-                f'      AuthType Basic',
-                f'      AuthName "Restricted Proxy"',
-                f'      AuthBasicProvider file',
+                '    <Proxy "*">',
+                '      AuthType Basic',
+                '      AuthName "Restricted Proxy"',
+                '      AuthBasicProvider file',
                 f'      AuthUserFile "{self._basic_passwords}"',
-                f'      Require user proxy',
-                f'    </Proxy>',
+                '      Require user proxy',
+                '    </Proxy>',
             ]
         else:
             return [
-                f'    <Proxy "*">',
-                f'      Require ip 127.0.0.1',
-                f'    </Proxy>',
+                '    <Proxy "*">',
+                '      Require ip 127.0.0.1',
+                '    </Proxy>',
             ]
 
     def _get_log_level(self):
@@ -413,44 +440,44 @@ class Httpd:
         lines = []
         if Httpd.MOD_CURLTEST is not None:
             lines.extend([
-                f'    Redirect 302 /data.json.302 /data.json',
-                f'    Redirect 301 /curltest/echo301 /curltest/echo',
-                f'    Redirect 302 /curltest/echo302 /curltest/echo',
-                f'    Redirect 303 /curltest/echo303 /curltest/echo',
-                f'    Redirect 307 /curltest/echo307 /curltest/echo',
-                f'    <Location /curltest/sslinfo>',
-                f'      SSLOptions StdEnvVars',
-                f'      SetHandler curltest-sslinfo',
-                f'    </Location>',
-                f'    <Location /curltest/echo>',
-                f'      SetHandler curltest-echo',
-                f'    </Location>',
-                f'    <Location /curltest/put>',
-                f'      SetHandler curltest-put',
-                f'    </Location>',
-                f'    <Location /curltest/tweak>',
-                f'      SetHandler curltest-tweak',
-                f'    </Location>',
-                f'    Redirect 302 /tweak /curltest/tweak',
-                f'    <Location /curltest/1_1>',
-                f'      SetHandler curltest-1_1-required',
-                f'    </Location>',
-                f'    <Location /curltest/shutdown_unclean>',
-                f'      SetHandler curltest-tweak',
-                f'      SetEnv force-response-1.0 1',
-                f'    </Location>',
-                f'    SetEnvIf Request_URI "/shutdown_unclean" ssl-unclean=1',
+                '    Redirect 302 /data.json.302 /data.json',
+                '    Redirect 301 /curltest/echo301 /curltest/echo',
+                '    Redirect 302 /curltest/echo302 /curltest/echo',
+                '    Redirect 303 /curltest/echo303 /curltest/echo',
+                '    Redirect 307 /curltest/echo307 /curltest/echo',
+                '    <Location /curltest/sslinfo>',
+                '      SSLOptions StdEnvVars',
+                '      SetHandler curltest-sslinfo',
+                '    </Location>',
+                '    <Location /curltest/echo>',
+                '      SetHandler curltest-echo',
+                '    </Location>',
+                '    <Location /curltest/put>',
+                '      SetHandler curltest-put',
+                '    </Location>',
+                '    <Location /curltest/tweak>',
+                '      SetHandler curltest-tweak',
+                '    </Location>',
+                '    Redirect 302 /tweak /curltest/tweak',
+                '    <Location /curltest/1_1>',
+                '      SetHandler curltest-1_1-required',
+                '    </Location>',
+                '    <Location /curltest/shutdown_unclean>',
+                '      SetHandler curltest-tweak',
+                '      SetEnv force-response-1.0 1',
+                '    </Location>',
+                '    SetEnvIf Request_URI "/shutdown_unclean" ssl-unclean=1',
             ])
         if self._auth_digest:
             lines.extend([
                 f'    <Directory {self.docs_dir}/restricted/digest>',
-                f'      AuthType Digest',
-                f'      AuthName "restricted area"',
+                '      AuthType Digest',
+                '      AuthName "restricted area"',
                 f'      AuthDigestDomain "https://{servername}"',
-                f'      AuthBasicProvider file',
+                '      AuthBasicProvider file',
                 f'      AuthUserFile "{self._digest_passwords}"',
-                f'      Require valid-user',
-                f'    </Directory>',
+                '      Require valid-user',
+                '    </Directory>',
 
             ])
         return lines
